#ifndef TAB_C
#define TAB_C
#include "termios.h"

int tab_ioctl(int fd) {                  /*** query terminal width */
    struct winsize win;                    /* xxx make more portable? */

    if (ioctl(fd, TIOCGWINSZ, &win) == -1)
        return 0;
    return win.ws_col;
}

int tab_term_cols() {                    /*** decide on *some* terminal size */
    int c, fd;

    if ((c = tab_ioctl(0)) || (c = tab_ioctl(1)) || (c = tab_ioctl(2)))
        return c;                          /* use a std fd */
    if ((fd = open("/dev/tty", O_RDONLY)) != -1) {
        c = tab_ioctl(fd);
        close(fd);
        if (c)
            return c;                      /* get & use ctty fd */
    }
    if (getenv("COLUMNS") && (c = atoi(getenv("COLUMNS"))))
        return c;                          /* env vars or finally defaults */
    return 80;                             /* if all else fails, use 80 */
}

int tab_printed_width(char *a) {         /*** compute width when printed */
    int n;                                 /* currently just elides \e[..m */

    for (n = 0; *a; a++, n++)
        if (a[0] == '\x1b' && a[1] == '[') {
            for (a++, a++, n--; *a && *a != 'm'; a++)
                /**/;
            if (*a != 'm')                 /* unterminated escape sequence */
                a--;                       /* backup so outer loop terminates */
        }
    return n;
}

int tab_sum(int *x, int n) {             /*** sum an array of ints */
    int i, sum = 0;

    for (i = 0; i < n; i++)
        sum += x[i];
    return sum;
}

int tab_max(int *x, int a, int b) {      /*** max over subarray of ints */
    int i, mx = -1000;
    for (i = a; i < b; i++)
        if (x[i] > mx)
            mx = x[i];
    return mx;
}

void
tab_rowcol(int*r, int*c, int n, int nc) {/*** handle ceil(n/nc) assignment */
    int div = n / nc;
    int mod = n % nc;
    *r = div + (mod != 0);
    *c = nc;
}

int *tab_layout(int *lens, int n,       /**** Determine layout */
                int W, int gap, int mx, int pd, int *Nrow, int *Ncol) {
    int *ws, nws, *ws_try, nws_try, nrow, ncol, i, c, excess, added;

    ws     = (int *) malloc((mx + 1) * sizeof *ws);
    ws_try = (int *) malloc((mx + 1) * sizeof *ws_try);
    tab_rowcol(&nrow, &ncol, n, 1);        /* 1 => W/(max(lens)+gap) */
    for (i = 0; i < ncol; i++)
        ws[i] = W / ncol;
    nws = ncol;
    while (nrow > 1 && ncol < mx) {       /** Basic fitting into columns */
        int nrow0 = nrow, a, b, w1;        /* add more cols until at least */
        while (nrow >= nrow0)              /* one row can be elided */
            tab_rowcol(&nrow, &ncol, n, ncol + 1);
        nws_try = 0;                       /* find widths needed by data */
        for (c = 0; c < ncol; c++) {
            a  = nrow * c;
            b  = (nrow * (c + 1) < n) ? nrow * (c + 1) : n;
            w1 = tab_max(lens, a, b) + gap;/* width = widest + gap */
            ws_try[nws_try++] = w1;
        }
        ws_try[nws_try - 1] -= gap;        /* last col needs no inter-col gap */
        if (tab_sum(ws_try, nws_try) > W)  /* out of room. DONE */
            break;
        for (i = 0; i < nws_try; i++)      /* new cols fit: copy config */
            ws[i] = ws_try[i];
        nws = nws_try;
    }
    tab_rowcol(&nrow, &ncol, n, nws);
    excess = W - tab_sum(ws, nws);         /* Distribute excess space -> gaps */
    c = 0;
    added = 0;
    while (excess > 0 && added < ncol * pd) {
        added++;                           /* add at most pd spc per col */
        ws[c]++;
        excess--;
        if (ncol > 1)
            c = (c + 1) % (ncol - 1);      /* cyclic index over columns */
    }
    *Nrow = nrow;
    *Ncol = ncol;
    free((void *)ws_try);
    return ws;
}

void tab_write(FILE *f, char **strs,    /**** write output lines */
               int *lens, int nstr, int *ws, int nrow, int ncol, char *pfx) {
    int i, r, c, maxspc = 4094;
    static char spaces[4096];

    memset(spaces, ' ', sizeof spaces - 2);
    spaces[sizeof spaces - 1] = '\0';
    for (r = 0; r < nrow; r++) {
        if (pfx && *pfx)
            fputs(pfx, f);
        for (c = 0; c < ncol; c++) {
            i = nrow * c + r;
            if (i >= nstr)
                break;
            fputs(strs[i], f);
            fputs(spaces + (maxspc - (ws[c] - lens[i])), f);
        }
        fputc('\n', f);
    }
}
#endif /* TAB_C */
